@@ -1,186 +1,45 @@
+/*****************************************************************************
+**
+** pnpdump.c
+*/
+
 #include <stdio.h>
 
 #ifdef __DJGPP__
 #undef REALTIME
-#include <inlines/pc.h>
-#include <dos.h>
 #else
 #ifndef _OS2_
 #include <getopt.h>
-#  if defined __GLIBC__ && __GLIBC__ >= 2
-#    include <sys/io.h>
-#  else
-#    ifdef _AXP_
-#       include <sys/io.h>
-#    else
-#       include <asm/io.h>
-#    endif
-#  endif												   /* __GLIBC__ */
-#else
-#include <sys/hw.h>
 #endif /* _OS2_ */
 #endif
 
-#include <unistd.h>
 #include <stdlib.h>
-#include <string.h>										   /* For strncpy */
+#include <string.h>		/* For strncpy */
 
 #define TAG_DEBUG 0
 #define DUMPADDR 0
 
+#include "iopl.h"
 #include "pnp.h"
 #include "resource.h"
-
-#define NUM_CARDS 128
-#define IDENT_LEN 9
-#define TMP_LEN 16
-#define LARGE_LEN 65536
-
-static char rcsid[] = "$Id: pnpdump.c,v 1.18 1999/02/14 22:47:18 fox Exp $";
+#include "res-access.h"
+#include "errenum.h"
+#include "callbacks.h"
+#include "release.h"
 
 #ifdef REALTIME
-/*
- * Code to set scheduler to Realtime, Round-Robin, so usleeps right etc
- */
-#if defined __GLIBC__ && __GLIBC__ >= 2
-#  include <sched.h>
-#  include <time.h>
-#else
-#  include <linux/unistd.h>
-#  include <linux/sched.h>
-#  include <sys/time.h>
+#include "realtime.h"
+long l_realtime_timeout = 5L;
+#endif
 
-#  ifdef NEEDSETSCHEDULER
-_syscall3(int, sched_setscheduler, pid_t, pid, int, policy, struct sched_param *, sched_p)
-#  endif
-#  ifdef NEEDNANOSLEEP
-_syscall2(int, nanosleep, struct timespec *, rqtp, struct timespec *, rmtp)
-#  endif
-#endif /* __GLIBC__ */
+#define LARGE_LEN 65536
 
-int realtimeok = 0;
-
-void
-setroundrobin(void)
-{
-	pid_t mypid = getpid();
-	struct sched_param sched_p =
-	{50};
-	/* Use the syscall, as my library not up todate */
-	if (sched_setscheduler(mypid, SCHED_RR, &sched_p) < 0)
-	{
-		perror("Couldn't set real-time scheduling, may be a bit slow");
-	}
-	else
-		realtimeok = 1;
-}
-
-/*
- * Use nanosleep for realtime delays
- */
-void
-delaynus(long del)
-{
-	struct timespec t;
-	t.tv_sec = 0;
-	t.tv_nsec = del * 1000;
-	/*
-	 * Need to handle case where binary for later kernel run on an earlier
-	 * one, which doesn't support nanosleep (emergency backup spare !)
-	 */
-	if (realtimeok)
-	{
-		if (nanosleep(&t, (struct timespec *) 0) < 0)
-		{
-			perror("nanosleep failed");
-			realtimeok = 0;
-		}
-	}
-	else
-		usleep(del);
-}
-
-#else
-#define delaynus(x) usleep(x)
-#undef NEEDSETSCHEDULER
-#undef NEEDNANOSLEEP
-#endif /* REALTIME */
-
-#ifdef _OS2_
-unsigned 
-inb(unsigned port)
-{
-	return (_inp8(port));
-}
-
-unsigned 
-inw(unsigned port)
-{
-	return (_inp16(port));
-}
-
-unsigned 
-inl(unsigned port)
-{
-	return (_inp32(port));
-}
-
-void 
-outb(unsigned value, unsigned port)
-{
-	_outp8(port, value);
-}
-
-/* void outw(unsigned port, unsigned value){_outp16(port, value);} */
-/* void outl(unsigned port, unsigned value){_outp32(port, value);} */
-unsigned 
-iopl(unsigned port)
-{
-	return (0);
-}
-
-unsigned 
-usleep(unsigned wait)
-{
-	return (_sleep2((wait + 990) / 1000));
-}
-
-#endif /* _OS2_ */
-
-#ifdef __DJGPP__
-unsigned 
-iopl(unsigned port)
-{
-	return (0);
-}
-
-unsigned 
-usleep(unsigned wait)
-{
-	return (delay((wait + 990) / 1000), 0);
-}
-
-#endif /* __DJGPP__ */
-
-#define ASSERT(condition)								   /* as nothing */
-
-typedef void (*resource_handler)
-     (FILE * output_file, struct resource * res, int selected);
-
-static void read_board_resources(struct resource *first_element,
-								 int dependent_clause,
-								 struct resource **result_array_ptr,
-								 int *count);
+static char rcsid[] = "$Id: pnpdump.c,v 1.21 1999/12/09 22:28:33 fox Exp $";
 
 static void print_resource(FILE * output_file,
-						   struct resource *res,
-						   int selected);
+			   struct resource *res,
+			   int selected);
 
-static void for_each_resource(struct resource *res,
-							  int num_resources,
-							  resource_handler handler,
-							  FILE * output_file,
-							  int selected);
 
 #ifndef __DJGPP__
 int do_script;
@@ -193,10 +52,6 @@ static void print_script(FILE * output_file,
 						 int selected);
 #endif /* __DJGPP__ */
 
-static void read_board(int csn, struct resource **res_list, int *list_len);
-
-static void unread_resource_data(char prev);
-
 /* Global state variables used for printing resource information. */
 unsigned long serialno = 0;
 char devid[8];											   /* Copy as devidstr
@@ -230,36 +85,110 @@ prepare_to_print(void)
 	indep = "";
 }
 
-unsigned char serial_identifier[NUM_CARDS + 1][IDENT_LEN];
-unsigned char csumdiff[NUM_CARDS + 1];
-
 unsigned char tmp[TMP_LEN];
 unsigned char large[LARGE_LEN];
 static char *devstring = 0;
 
 int showmasks = 0;
 int do_autoconfig = 0;
-int do_fullreset = 0;
 
-int csum = 0;
-int backspaced = 0;
-char backspaced_char;
-int read_port = MIN_READ_ADDR;
-int boards_found = 0;
-
-int do_isolate(void);
-void initiate(void);
 void dumpdata(int);
-void read_idents(void);
 
 static void usage(char *program_name);
 
 int do_dumpregs = 0;
-int do_ignorecsum = 0;
 void dumpregs(int);
 
 char *devidstr(unsigned char, unsigned char, unsigned char, unsigned char);
 
+/****************************************************************************/
+/* Callbacks needed by the libisapnp procedures                             */
+/****************************************************************************/
+
+char *st_error_messages[] = {
+#define ISAPNP_ERR_CODE(code,msg) msg,
+#include "errcodes.h"
+};
+
+/* Forward declarations - this way I am making sure that the typedefs and the
+** actual declarations (below) match each other.
+*/
+fatal_error     pnpdump_fatal_error_callback;
+non_fatal_error pnpdump_non_fatal_error_callback;
+progress_report pnpdump_progress_report_callback;
+
+/* The callbacks themselves */
+void
+pnpdump_fatal_error_callback(int in_isapnp_error)
+{
+  if (in_isapnp_error >= ISAPNP_E_LAST_ERROR_CODE) {
+    fprintf(stderr, "Unknown fatal error - error code is %d.\n",
+	    in_isapnp_error);
+  }
+  else {
+    fprintf(stderr,"%s\n",st_error_messages[in_isapnp_error]);
+  }
+  exit(1);
+}
+
+void
+pnpdump_non_fatal_error_callback(int in_errno, int in_isapnp_error)
+{
+  if (in_isapnp_error >= ISAPNP_E_LAST_ERROR_CODE) {
+    fprintf(stderr, "Unknown fatal error - error code is %d: %s\n",
+	    in_isapnp_error, strerror(in_errno));
+  }
+  else if (ISAPNP_E_PRINT_PROGRESS_REPORT_BUF == in_isapnp_error) {
+    fprintf(stderr, progress_report_buf);
+  }
+  else {
+    if (0 == in_errno) {
+      fprintf(stderr,"%s\n", st_error_messages[in_isapnp_error]);
+    }
+    else {
+		perror(st_error_messages[in_isapnp_error]);
+    }
+  }
+}
+
+void
+pnpdump_progress_report_callback(const char *in_msg)
+{
+  /* Prints the progress report to stdout.
+  ** In pnpdump.c it is used to actually write the isapnp.conf
+  ** configuration file.
+  */
+  printf("%s", in_msg);
+}
+
+/****************************************************************************/
+
+static void
+banner(FILE *fp, char *prefix)
+{
+	fprintf(fp, 
+			"%s%s\n"
+			"%sRelease %s (library %s)\n"
+			"%s\n"
+			"%sThis is free software, see the sources for details.\n"
+			"%sThis software has NO WARRANTY, use at your OWN RISK\n"
+			"%s\n"
+			"%sFor details of the output file format, see isapnp.conf(5)\n"
+			"%s\n"
+			"%sFor latest information and FAQ on isapnp and pnpdump see:\n"
+			"%shttp://www.roestock.demon.co.uk/isapnptools/\n"
+			"%s\n"
+			"%sCompiler flags: %s\n"
+			"%s(for   library: %s)\n"
+			"%s\n",
+			prefix, rcsid,
+			prefix, ISAPNPTOOLSVER, libtoolsver,
+			prefix, prefix, prefix, prefix, prefix, prefix, prefix, prefix, prefix,
+			prefix, ISAPNPTOOLSCOMPILERFLAGS,
+			prefix, libcompilerflags,
+			prefix);
+}
+
 /*
  * Single argument is initial readport
  * Two arguments is no-of-boards and the readport to use
@@ -272,20 +201,28 @@ main(int argc, char **argv)
 	struct resource *resources;
 	int resource_count;
 	int alloc_result;
+
+	/* Process command line options */
 #ifndef __DJGPP__
 	const struct option longopts[] =
 	{
 		{"masks", no_argument, NULL, 'm'},
 		{"config", no_argument, NULL, 'c'},
+		{"help", no_argument, NULL, 'h'},
 		{"script", optional_argument, NULL, 's'},
 		{"reset", no_argument, NULL, 'r'},
 		{"dumpregs", no_argument, NULL, 'd'},
+		{"debug-isolate", no_argument, NULL, 'D'},
 		{"ignorecsum", no_argument, NULL, 'i'},
+#ifdef REALTIME
+		{"max-realtime", required_argument, NULL, 't'},
+#endif
+		{"version", no_argument, NULL, 'v'},
 		{0, 0, 0, 0},
 	};
 	int opt;
 
-	while ((opt = getopt_long(argc, argv, "cms:rdi", longopts, NULL)) != EOF)
+	while ((opt = getopt_long(argc, argv, "cms:rdit:vDh", longopts, NULL)) != EOF)
 	{
 		switch (opt)
 		{
@@ -305,17 +242,36 @@ main(int argc, char **argv)
 		case 'd':
 			do_dumpregs = 1;
 			break;
+		case 'D':
+			debug_isolate = 1;
+			break;
+		case 'h':
+			usage(program);
+			break;
 		case 'i':
 			do_ignorecsum = 1;
 			break;
+#ifdef REALTIME
+		case 't':
+			l_realtime_timeout = atol(optarg);
+			break;
+#else
+		case 't':
+			fprintf(stderr, "Realtime support not compiled in - option ignored\n");
+			break;
+#endif
+		case 'v':
+			fprintf(stderr, "Version: pnpdump from %s\n", ISAPNPTOOLSVER);
+			exit(0);
+			break;
 		case '?':
-#if 0
+#if 1
 			fprintf(stderr, "unrecognized option.\n");
 #endif
 			usage(program);
 			return 1;
 		case ':':
-#if 0
+#if 1
 			fprintf(stderr, "missing parameter.\n");
 #endif
 			usage(program);
@@ -333,7 +289,11 @@ main(int argc, char **argv)
 #else
 	while (argc > 1)
 	{
-		if (strcmp(argv[1], "--reset") == 0)
+		if (strcmp(argv[1], "--help") == 0)
+		{
+			usage(program);
+		}
+		else if (strcmp(argv[1], "--reset") == 0)
 		{
 			do_fullreset = 1;
 		}
@@ -353,6 +313,14 @@ main(int argc, char **argv)
 		{
 			do_ignorecsum = 1;
 		}
+		else if (strcmp(argv[1], "--debug-isolate") == 0)
+		{
+			debug_isolate = 1;
+		}
+		else if (strcmp(argv[1], "--version") == 0)
+		{
+			fprintf(stderr, "Version: pnpdump from %s\n", ISAPNPTOOLSVER);
+		}
 		else
 		{
 			break;
@@ -362,6 +330,12 @@ main(int argc, char **argv)
 	}
 #endif /* __DJGPP__ */
 
+	/* Initialize callbacks */
+	callbacks_init(pnpdump_fatal_error_callback,
+		       pnpdump_non_fatal_error_callback,
+		       pnpdump_progress_report_callback);
+
+	/* Process arguments */
 	if ((argc < 1) || (argc > 3))
 	{
 		usage(program);
@@ -389,50 +363,16 @@ main(int argc, char **argv)
 		}
 		read_port |= 3;
 	}
-#ifdef _AXP_
-	/* ALPHA only has ioperm, apparently, so cover all with one permission */
-	if (ioperm(MIN_READ_ADDR, WRITEDATA_ADDR - MIN_READ_ADDR + 1, 1))
-#else
-	/*
-	 * Have to get unrestricted access to io ports, as WRITE_DATA port >
-	 * 0x3ff
-	 */
-	if (iopl(3) < 0)
-#endif
-	{
-		perror("Unable to get io permission for WRITE_DATA");
-		exit(1);
+	if (acquire_pnp_io_privileges() != 0) {
+	  perror("Unable to get io permission for WRITE_DATA");
+	  exit(1);
 	}
-#ifdef REALTIME
-	setroundrobin();
-#endif /* REALTIME */
-	printf("# %s\n# This is free software, see the sources for details.\n"
-		   "# This software has NO WARRANTY, use at your OWN RISK\n"
-		   "#\n"
-		   "# For details of this file format, see isapnp.conf(5)\n"
-		   "#\n"
-		   "# For latest information and FAQ on isapnp and pnpdump see:\n"
-		   "# http://www.roestock.demon.co.uk/isapnptools/\n"
-		   "#\n"
-		   "# Compiler flags:", rcsid);
-#ifdef _AXP_
-	printf(" -D_AXP_");
-#endif /* _AXP_ */
-#ifdef REALTIME
-	printf(" -DREALTIME");
-#endif /* REALTIME */
-#ifdef NEEDSETSCHEDULER
-	printf(" -DNEEDSETSCHEDULER");
-#endif /* NEEDSETSCHEDULER */
-#ifdef NEEDNANOSLEEP
-	printf(" -DNEEDNANOSLEEP");
-#endif /* NEEDNANOSLEEP */
-#ifdef ABORT_ONRESERR
-	printf(" -DABORT_ONRESERR");
-#endif /* ABORT_ONRESERR */
-	printf("\n#\n");
+	banner(stdout, "# ");
 	/* Have to do this anyway to check readport for clashes */
 	alloc_system_resources();
+#ifdef REALTIME
+	setroundrobin(l_realtime_timeout);
+#endif /* REALTIME */
 	/* Now get board info */
 	initiate();
 	resources = NULL;
@@ -447,6 +387,10 @@ main(int argc, char **argv)
 		if(do_dumpregs)
 			dumpregs(i);
 	}
+#ifdef REALTIME
+	/* No longer need real-time scheduling */
+	normal_sched();
+#endif /* REALTIME */
 	if (do_autoconfig)
 	{
 		alloc_result = alloc_resources(resources, resource_count, NULL, 0);
@@ -462,11 +406,11 @@ main(int argc, char **argv)
 	printf("# Returns all cards to the \"Wait for Key\" state\n");
 	printf("(WAITFORKEY)\n");
 	/* Release resources */
-#ifdef _AXP_
-	ioperm(MIN_READ_ADDR, WRITEDATA_ADDR - MIN_READ_ADDR + 1, 0);
-#else
-	(void) iopl(0);
-#endif
+	if (relinquish_pnp_io_privileges() != 0) {
+	  perror("Unable to release io permission for WRITE_DATA");
+	  exit(1);
+	}
+
 
 #ifndef __DJGPP__
 	if (do_script)
@@ -512,265 +456,38 @@ main(int argc, char **argv)
 static void 
 usage(char *program_name)
 {
-	fprintf(stderr, "\n%s\nThis is free software, see the sources for details.\n"
-			"For latest information, see "
-			"http://www.roestock.demon.co.uk/isapnptools/\n"
-			"This software has NO WARRANTY, use at your OWN RISK\n\n"
+	banner(stderr, "");
+	fprintf(stderr,
+			"Usage: %s [OPTIONS] [[devs] readport]\n\n"
 #ifndef __DJGPP__
-			"Usage: %s [--masks] [--config] [--reset] [--script=outputfile] [--dumpregs] [--ignorecsum]\n",
-#else
-			"Usage: %s [--masks] [--config] [--reset] [--dumpregs] [--ignorecsum]\n",
-#endif
-			rcsid, program_name);
-}
-
-/*
- * Wait for resource data byte
- *
- * Return true if error occurred (timeout)
- */
-int
-statuswait(void)
-{
+			"   -c, --config          uncomment settings to which the devices can be set\n"
+			"   -d, --dumpregs        dump standard configuration registers for each board\n"
+			"   -D, --debug-isolate   output loads of extra information on isolation process\n"
+			"   -h, --help            show a help summary to stderr\n"
+			"   -i, --ignorecsum      ignore checksum errors when choosing the readport address\n"
+			"   -r, --reset           carry out a full configuration reset\n"
+			"   -m, --masks           list acceptable interrupts and DMA channels as bitmasks\n"
 #ifdef REALTIME
-#define TIMEOUTLOOPS 100
-	int to;												   /* For timeout */
-	/*
-	 * Try for up to 1ms
-	 */
-	for (to = 0; to < TIMEOUTLOOPS; to++)
-	{
-		if (STATUS & 1)
-			break;
-		delaynus(10L);
-	}
-	if (to >= TIMEOUTLOOPS)
-	{
-		fprintf(stderr, "Timeout attempting to read resource data - is READPORT correct ?\n");
-		return (1);
-	}
-#else /* !REALTIME */
-#if defined _OS2_ || defined __DJGPP__
-#define TIMEOUTLOOPS 2
-	int to;												   /* For timeout */
-	/*
-	 * Try for up to 2ms
-	 */
-	for (to = 0; to < TIMEOUTLOOPS; to++)
-	{
-		usleep(1000);
-		if (STATUS & 1)
-			break;
-	}
-	if (to >= TIMEOUTLOOPS)
-	{
-		fprintf(stderr, "Timeout attempting to read resource data - is READPORT correct ?\n");
-		return (1);
-	}
-#else /* !(OS2 or DJGPP) */
-	/*
-	 * Infinite loop potentially, but if we usleep, we may lose 10ms
-	 */
-	while (!(STATUS & 1))
-		;
-#endif /* !(OS2 or DJGPP) */
-#endif /* !REALTIME */
-	return (0);
-}
-
-int port;
-
-void
-sendkey(void)
-{
-	static char initdata[INIT_LENGTH] = INITDATA;
-	int i;
-	ADDRESS(0);
-	ADDRESS(0);
-	for (i = 0; i < INIT_LENGTH; i++)
-	{
-		ADDRESS(initdata[i]);
-	}
-}
-
-void
-initiate(void)
-{
-	int i;
-	if (!boards_found)
-	{
-		/* All cards now isolated, read the first one */
-		for (port = read_port; port <= MAX_READ_ADDR; port += READ_ADDR_STEP)
-		{
-			/* Check it doesn't clash with anything */
-			if(!allocate_resource(IOport_TAG, port, 1, "readport"))
-			{
-				continue;
-			}
-			deallocate_resource(IOport_TAG, port, 1);
-			/* Make sure all cards are in Wait for key */
-			CONFIGCONTROL;
-			WRITE_DATA(CONFIG_WAIT_FOR_KEY);
-			/* Make them listen */
-			delaynus(2000L);
-			sendkey();
-			/* Reset the cards */
-			CONFIGCONTROL;
-			WRITE_DATA(CONFIG_RESET_CSN);
-			if(do_fullreset)
-			{
-				CONFIGCONTROL;
-				WRITE_DATA(CONFIG_RESET);
-				delaynus(2000L);
-			}
-			printf("# Trying port address %04x\n", port);
-			if (do_isolate())
-				break;
-		}
-		if (port > MAX_READ_ADDR)
-		{
-			printf("# No boards found\n");
-			exit(0);
-		}
-		printf("# Board %d has serial identifier", boards_found);
-		for (i = IDENT_LEN; i--;)
-			printf(" %02x", serial_identifier[boards_found][i]);
-		printf("\n");
-		while (do_isolate())
-		{
-			printf("# Board %d has serial identifier", boards_found);
-			for (i = IDENT_LEN; i--;)
-				printf(" %02x", serial_identifier[boards_found][i]);
-			printf("\n");
-		}
-		printf("\n# (DEBUG)\n(READPORT 0x%04x)\n", port);
-		if(do_ignorecsum)
-			printf("(IGNORECRC)\n");
-		if(do_fullreset)
-			printf("(ISOLATE CLEAR)\n");
-		else
-			printf("(ISOLATE PRESERVE)\n");
-		printf("(IDENTIFY *)\n");
-	}
-	else
-	{
-		CONFIGCONTROL;
-		WRITE_DATA(CONFIG_WAIT_FOR_KEY);
-		sendkey();
-		read_idents();
-		printf("\n# (DEBUG)\n(READPORT 0x%04x)\n(CSN %d)\n(IDENTIFY *)\n", read_port, boards_found);
-	}
-	printf("(VERBOSITY 2)\n(CONFLICT (IO FATAL)(IRQ FATAL)(DMA FATAL)(MEM FATAL)) # or WARNING\n\n");
-}
-
-void
-read_idents(void)
-{
-	int csn = 0;
-	int i;
-	for (csn = 1; csn <= boards_found; csn++)
-	{
-		Wake(csn);
-		for (i = 0; i < IDENT_LEN; i++)
-		{
-			if (statuswait())
-				return;
-			serial_identifier[csn][i] = RESOURCEDATA;
-		}
-		printf("# Board %d has serial identifier", csn);
-		for (i = IDENT_LEN; i--;)
-			printf(" %02x", serial_identifier[csn][i]);
-		printf("\n");
-		if (serial_identifier[csn][0] == 0xff)
-			boards_found = csn - 1;
-	}
-}
-
-int
-do_isolate(void)
-{
-	unsigned char c1, c2;
-	int i;
-	int index;
-	int newbit;
-	int goodaddress = 0;
-	if (boards_found >= NUM_CARDS)
-	{
-		fprintf(stderr, "Too many boards found, recompile with NUM_CARDS bigger\n");
-		return 0;
-	}
-	csum = 0x6a;
-	/* Assume we will find one */
-	boards_found++;
-	Wake(0);
-	SetRdPort(port);
-	delaynus(1000L);
-	SERIALISOLATION;
-	delaynus(2000L);
-	for (index = 0; index < IDENT_LEN - 1; index++)
-	{
-		for (i = 0; i < 8; i++)
-		{
-			newbit = 0x00;
-			delaynus(250L);
-			c1 = READ_DATA;
-			delaynus(250L);
-			c2 = READ_DATA;
-			if (c1 == 0x55)
-			{
-				if (c2 == 0xAA)
-				{
-					goodaddress = 1;
-					newbit = 0x80;
-				}
-				else
-				{
-					goodaddress = 0; 
-				}
-			}
-			/* printf("\n %02x %02x - bit %02x, goodaddress %d",c1,c2, newbit, goodaddress); */
-			serial_identifier[boards_found][index] >>= 1;
-			serial_identifier[boards_found][index] |= newbit;
-			/* Update checksum */
-			if (((csum >> 1) ^ csum) & 1)
-				newbit ^= 0x80;
-			csum >>= 1;
-			csum |= newbit;
-		}
-		/* printf(" *** %02x \n", serial_identifier[boards_found][index]); */
-	}
-	/* printf("computed csum is %02x", csum); */
-	for (i = 0; i < 8; i++)
-	{
-		newbit = 0x00;
-		delaynus(250L);
-		c1 = READ_DATA;
-		delaynus(250L);
-		c2 = READ_DATA;
-		if (c1 == 0x55)
-		{
-			if (c2 == 0xAA)
-			{
-				goodaddress = 1;
-				newbit = 0x80;
-			}
-		}
-		/* printf("\n %02x %02x - bit %02x, goodaddress %d",c1,c2, newbit, goodaddress); */
-		serial_identifier[boards_found][index] >>= 1;
-		serial_identifier[boards_found][index] |= newbit;
-	}
-	/* printf(" *** csum *** %02x\n", serial_identifier[boards_found][index]); */
-	if (goodaddress && (do_ignorecsum || (csum == serial_identifier[boards_found][index])))
-	{
-		CARDSELECTNUMBER;
-		WRITE_DATA(boards_found);
-		if((csumdiff[boards_found] = (csum - serial_identifier[boards_found][index])))
-			printf("# WARNING: serial identifier mismatch for board %d: expected 0x%02X, got 0x%02X\n", boards_found, serial_identifier[boards_found][index], csum);
-		return (1);
-	}
-	/* We didn't find one */
-	boards_found--;
-	return (0);
+			" -t t, --max-realtime=t  maximum real-time priority execution is t seconds\n"
+#endif
+			" -s scr, --script[=scr]  write a shell script to the specified file\n"
+			"                         no script pipes output to a shell\n"
+			"   -v, --version         print the isapnptools version number on stderr\n"
+#else
+			" 	--config          uncomment settings to which the devices can be set\n"
+			" 	--dumpregs        dump standard configuration registers for each board\n"
+			" 	--debug-isolate   output loads of extra information on isolation process\n"
+			" 	--help            show a help summary to stderr\n"
+			" 	--ignorecsum      ignore checksum errors when choosing the readport address\n"
+			" 	--reset           carry out a full configuration reset\n"
+			" 	--masks           list acceptable interrupts and DMA channels as bitmasks\n"
+			" 	--version         print the isapnptools version number on stderr\n"
+#endif
+			"\n"
+			"   devs      is the number of PnP cards that the BIOS has found\n"
+			"   readport  is the address of the readport to use for Plug-And-Play access\n"
+			, program_name);
+	exit(0);
 }
 
 char *
@@ -838,237 +555,6 @@ showmask(unsigned long mask)
 	}
 }
 
-static void 
-unread_resource_data(char prev)
-{
-	csum = (csum - prev) & 0xFF;
-	backspaced = 1;
-	backspaced_char = prev;
-}
-
-/* returns 0 on success, nonzero on failure. */
-
-static inline int
-read_resource_data(unsigned char *result)
-{
-	if (backspaced)
-	{
-		*result = backspaced_char;
-		backspaced = 0;
-		return 0;
-	}
-	if (statuswait())
-		return 1;
-	*result = RESOURCEDATA;
-	csum = (csum + *result) & 0xFF;
-	return 0;
-}
-
-static int
-read_one_resource(struct resource *res)
-{
-	int i;
-	if (read_resource_data(&res->tag) != 0)
-		return 1;
-	if (res->tag & 0x80)
-	{
-		/* Large item */
-		res->type = res->tag;
-		for (i = 1; i <= 2; i++)
-		{
-			if (read_resource_data(&tmp[i]) != 0)
-				return 1;
-		}
-		res->len = (tmp[2] << 8) | tmp[1];
-	}
-	else
-	{
-		/* Small item */
-		res->type = (res->tag >> 3) & 0x0f;
-		res->len = res->tag & 7;
-	}
-	if ((res->data = malloc(res->len * sizeof(unsigned char))) == NULL)
-	{
-		fprintf(stderr, "read_one_resource: memory allocation failure.\n");
-		exit(1);
-	}
-	for (i = 0; i < res->len; i++)
-	{
-		if (read_resource_data(&res->data[i]) != 0)
-			return 1;
-	}
-
-#if TAG_DEBUG
-	printf("# %s %02X Tag %02X[%02X] :",
-		   (res->tag & 0x80) ? "L" : "S", res->tag, res->type, res->len);
-	for (i = 0; i <= res->len; i++)
-	{
-		printf(" %02x", res->data[i]);
-	}
-	printf("\n");
-#endif
-	return 0;
-}
-
-static void
-fill_in_resource(struct resource *res)
-{
-	res->mask = ~0;
-	res->start = 0;				/* Set some defaults */
-	res->size = 1;
-	res->step = 1;
-	switch (res->type)
-	{
-	case IRQ_TAG:
-		res->end = 16;
-		res->mask = res->data[0] | (res->data[1] << 8);
-		break;
-	case DMA_TAG:
-		res->end = 8;
-		res->mask = res->data[0];
-		break;
-	case IOport_TAG:
-		res->start = (res->data[2] << 8) | res->data[1];
-		res->end = ((res->data[4] << 8) | res->data[3]) + 1;
-		res->step = res->data[5];
-		res->size = res->data[6];
-		break;
-	case FixedIO_TAG:
-		res->start = (res->data[1] << 8) | res->data[0];
-		res->size = res->data[2];
-		res->end = res->start + 1;
-		break;
-	case MemRange_TAG:
-		res->start = (((long) res->data[2] << 8) | res->data[1]) * 256;
-		res->end = (((long) res->data[4] << 8) | res->data[3]) * 256 + 1;
-		res->step = (res->data[6] << 8) | res->data[5];
-		res->size = (((long) res->data[8] << 8) | res->data[7]) * 256;
-		break;
-	case Mem32Range_TAG:
-		res->start = res->end = 0;
-		/* STUB */
-		break;
-	case FixedMem32Range_TAG:
-		res->start = res->end = 0;
-		/* STUB */
-		break;
-	default:
-		break;
-	}
-	/* Set initial value. */
-	for (res->value = res->start; res->value < res->end; res->value += res->step)
-	{
-		if ((1 << (res->value & 0x1F)) & res->mask)
-			break;
-	}
-}
-
-static void
-read_alternative(struct alternative *alt, struct resource *start_element)
-{
-	ASSERT(start_element->type == StartDep_TAG);
-	switch (start_element->len)
-	{
-	case 0:
-		alt->priority = 1;
-		break;					/* 1 == "acceptable" */
-	case 1:
-		alt->priority = start_element->data[0];
-		break;
-	default:
-		fprintf(stderr, "read_alternative: Illegal StartDep_TAG length %d.\n", start_element->len);
-		break;
-	}
-	alt->resources = NULL;
-	alt->len = 0;
-	read_board_resources(start_element, 1, &alt->resources, &alt->len);
-	*start_element = alt->resources[alt->len - 1];
-	(alt->len)--;
-}
-
-static void
-read_board_resources(struct resource *first_element,
-					 int dependent_clause,
-					 struct resource **result_array_ptr,
-					 int *count)
-{
-	struct resource *result_array;
-	int elts;
-
-	result_array = *result_array_ptr;
-	elts = *count;
-	if (!result_array)
-	{
-		result_array = malloc(sizeof(struct resource));
-		*count = elts = 0;
-	}
-	if (!result_array)
-	{
-		fprintf(stderr,
-				"read_board_resources: malloc of result_array failed.\n");
-		exit(1);
-	}
-	if (first_element != NULL)
-	{
-		elts++;
-		if ((result_array =
-			 realloc(result_array, elts * sizeof(struct resource))) == NULL)
-		{
-			fprintf(stderr, "read_board_resources: realloc of result_array failed.\n");
-			exit(1);
-		}
-		result_array[elts - 1] = *first_element;
-	}
-	do
-	{
-		elts++;
-		result_array = realloc(result_array, elts * sizeof(struct resource));
-		if (!result_array)
-		{
-			fprintf(stderr,
-				 "read_board_resources: realloc of result_array failed.\n");
-			exit(1);
-		}
-		memset(tmp, 0, sizeof(tmp));
-		read_one_resource(&result_array[elts - 1]);
-		if (result_array[elts - 1].type == StartDep_TAG && !dependent_clause)
-		{
-			int num_alts = 1;
-			struct alternative *alts;
-			alts = malloc(sizeof(struct alternative));
-			do
-			{
-				if (num_alts != 1)
-				{
-					alts = realloc(alts, num_alts * sizeof(struct alternative));
-				}
-				if (!alts)
-				{
-					fprintf(stderr, "read_board_resources: memory allocation for alternatives failed.\n");
-					exit(1);
-				}
-				read_alternative(&alts[num_alts - 1], &result_array[elts - 1]);
-				num_alts++;
-			}
-			while (result_array[elts - 1].type == StartDep_TAG);
-			result_array[elts - 1].start = 0;
-			result_array[elts - 1].end = num_alts - 1;
-			result_array[elts - 1].mask = ~0;
-			result_array[elts - 1].step = 1;
-			result_array[elts - 1].alternatives = alts;
-		}
-		else
-		{
-			fill_in_resource(&result_array[elts - 1]);
-		}
-	}
-	while (result_array[elts - 1].type != End_TAG &&
-		   (!dependent_clause ||
-			(result_array[elts - 1].type != EndDep_TAG &&
-			 result_array[elts - 1].type != StartDep_TAG)));
-	*count = elts;
-	*result_array_ptr = result_array;
-}
 
 static int IORangeCheck = 0;
 
@@ -1146,7 +632,7 @@ print_resource(FILE * output_file, struct resource *res, int selected)
 			int reg;
 			static char *regfns[8] =
 			{"Device capable of taking part in boot process",
-			 "Device support I/O range check register",
+			 "Device supports I/O range check register",
 			 "Device supports reserved register @ 0x32",
 			 "Device supports reserved register @ 0x33",
 			 "Device supports reserved register @ 0x34",
@@ -1237,6 +723,10 @@ print_resource(FILE * output_file, struct resource *res, int selected)
 			if ((len >= 2) && (res->data[0] || res->data[1]))
 			{
 				printf("# %sIRQ", indep);
+				if(res->errflags & RES_ERR_NO_IRQ)
+					printf("# %s%s*** Bad resource data: No IRQ specified\n", indep, indep);
+				if(res->errflags & RES_ERR_IRQ2)
+					printf("# %s%s*** Bad resource data (Clarifications 4.6.2): IRQ 2 invalid, changing to 9\n", indep, indep);
 				showmask(res->mask);
 				printf(".\n");
 				firstirq = res->value;
@@ -1392,6 +882,8 @@ print_resource(FILE * output_file, struct resource *res, int selected)
 			printf("# %s%sMinimum IO base address 0x%04lx\n", indep, indep, res->start);
 			printf("# %s%sMaximum IO base address 0x%04lx\n", indep, indep, res->end - 1);
 			printf("# %s%sIO base alignment %ld bytes\n", indep, indep, res->step);
+			if(res->errflags & RES_ERR_NO_STEP)
+				printf("# %s%s*** Bad resource data: Base alignment 0 - changed to 1\n", indep, indep);
 			printf("# %s%sNumber of IO addresses required: %d\n", indep, indep, res->data[6]);
 #if DUMPADDR
 			for (i = ((res->data[3] << 8) + res->data[2]); i <= ((res->data[5] << 8) + res->data[4]); i += res->data[5])
@@ -1744,7 +1236,11 @@ print_script(FILE * output_file, struct resource *res, int selected)
 		break;
 	case End_TAG:
 		{
-			fprintf(output_file, "pnp_isa_configure\n\n");
+			if (nestdepth)
+			{
+				fprintf(output_file, "pnp_isa_configure\n\n");
+				nestdepth = 0;
+			}
 			break;
 		}
 	case Mem32Range_TAG:
@@ -1772,127 +1268,6 @@ print_script(FILE * output_file, struct resource *res, int selected)
 }
 #endif /* __DJGPP__ */
 
-static void 
-for_each_resource(struct resource *res,
-				  int num_resources,
-				  resource_handler handler,
-				  FILE * output_file,
-				  int selected)
-{
-	int i;
-	for (i = 0; i < num_resources; i++)
-	{
-		(*handler) (output_file, &res[i], selected);
-	}
-}
-
-static void
-read_board(int csn, struct resource **res_list, int *list_len)
-{
-	int i;
-	struct resource *res_array;
-	int array_len;
-
-	res_array = *res_list;
-	array_len = *list_len;
-	if (res_array == NULL || array_len == 0)
-	{
-		res_array = malloc(sizeof(struct resource));
-		array_len = 1;
-	}
-	else
-	{
-		array_len++;
-		res_array = realloc(res_array,
-							array_len * sizeof(struct resource));
-	}
-	if (!res_array)
-	{
-		fprintf(stderr,
-				"read_board: failed to allocate memory for res_array.\n");
-		exit(1);
-	}
-
-	res_array[array_len - 1].type = NewBoard_PSEUDOTAG;
-	res_array[array_len - 1].start = csn;
-	res_array[array_len - 1].end = csn + 1;
-	if ((res_array[array_len - 1].data = malloc(IDENT_LEN)) == NULL)
-	{
-		fprintf(stderr,
-		   "read_board: failed to allocate memory for board identifier.\n");
-		exit(1);
-	}
-
-	Wake(csn);
-	/*
-	 * Check for broken cards that don't reset their resource pointer
-	 * properly
-	 */
-	/* Get the first byte */
-	if (read_resource_data(&res_array[array_len - 1].data[0]) != 0)
-		return;
-	/* Just check the first byte, if these match, or the serial identifier had a checksum error, we assume it's ok */
-	if((!csumdiff[csn])&&(res_array[array_len - 1].data[0] != serial_identifier[csn][0]))
-	{
-		/*
-		 * Assume the card is broken and this is the start of the resource
-		 * data.
-		 */
-		unread_resource_data(res_array[array_len - 1].data[0]);
-		res_array[array_len - 1].len = 1;
-		goto broken;			/* Wouldn't need this if people read the spec */
-	}
-	res_array[array_len - 1].len = IDENT_LEN;
-	/*
-	 * Read resource data past serial identifier - As it should be spec
-	 * section 4.5 para 2
-	 */
-	for (i = 1; i < IDENT_LEN; i++)
-	{
-		if (read_resource_data(&res_array[array_len - 1].data[i]) != 0)
-			return;
-	}
-	if(csumdiff[csn])
-	{
-		/* Original serial identifier had checksum error, so make the serial
-		 * identifier what we read from the resource data, as that's what
-		 * isapnp will see from an (identify *), assuming checksum ok
-		 */
-		int data;
-		int newbit;
-		int j;
-		csum = 0x6a;
-		for (i = 0; i < IDENT_LEN - 1; i++)
-		{
-			data = res_array[array_len - 1].data[i];
-			for(j = 0; j < 8; j++)
-			{
-				newbit = (data & 1) << 7;
-				data >>= 1;
-				if(((csum >> 1) ^ csum) & 1)
-					newbit ^= 0x80;
-				csum >>= 1;
-				csum |= newbit;
-			}
-		}
-		if(csum == res_array[array_len - 1].data[IDENT_LEN - 1])
-		{
-			printf("# WARNING: Making Board %d serial identifier the resource data version\n", csn);
-			for (i = 0; i < IDENT_LEN; i++)
-			{
-				serial_identifier[csn][i] = res_array[array_len - 1].data[i];
-			}
-		}
-		else
-			printf("# WARNING: Board %d resource data identifier has checksum error too\n", csn);
-	}
-	/* Now for the actual resource data */
-	csum = 0;
-  broken:
-	*res_list = res_array;
-	*list_len = array_len;
-	read_board_resources(NULL, 0, res_list, list_len);
-}
 
 #if 0
 void
@@ -2017,6 +1392,8 @@ dumpregs(int csn)
 	{
 		printf("# Vendor Id %s, Serial Number %lu, checksum 0x%02X.\n", devid, serialno, serial_identifier[csn][8]);
 	}
+	ADDRESS(0x06);
+	printf("#\n# Card Select Number register [0x06]: %d\n", READ_DATA);
 	printf("#\n# Vendor defined card level registers 0x20..2f:");
 	for (addr = 0x20; addr < 0x30; addr++)
 	{
@@ -2031,6 +1408,28 @@ dumpregs(int csn)
 		if (READ_DATA != logical_device)
 			break;
 		printf("#\n# Logical device %d\n", logical_device);
+		ADDRESS(0x30);
+		printf("# Activate       register [0x30]: %s\n", READ_DATA ? "* Activated *" : "Inactive");
+		ADDRESS(0x31);
+		printf("# IO range check register [0x31]: ");
+		tmp[0] = READ_DATA;
+		if(tmp[0] & 2)
+			printf("IO range check active, reads return %02x", tmp[0] & 1 ? 0x55 : 0xAA);
+		else
+			printf("Inactive");
+		printf("\n# Logical device Control reserved reg 0x32..37:");
+		for (addr = 0x32; addr < 0x38; addr++)
+		{
+			ADDRESS(addr);
+			printf(" %02x", READ_DATA);
+		}
+		printf("\n# Logical device Control Vendor defnd 0x38..40:");
+		for (addr = 0x38; addr < 0x40; addr++)
+		{
+			ADDRESS(addr);
+			printf(" %02x", READ_DATA);
+		}
+		putchar('\n');
 		for (addr = 0x40, desc = 0; addr < 0x60; addr += 8, desc++)
 		{
 			printf("# 24 bit Memory descriptor %d at %02x..%02x: ", desc, addr, addr + 4);
@@ -2086,3 +1485,5 @@ dumpregs(int csn)
 	}
 	putchar('\n');
 }
+
+/* End of pnpdump.c */
